// CpkFile.cpp : implementation file
//

#include "stdafx.h"
#include <vfw.h>
#include "CpkFile.h"

void SwapEndian( DWORD& dw)
{
	BYTE	temp1[4],
			temp2[4];

	*(DWORD*)&temp1 = dw;
	temp2[0] = temp1[3];
	temp2[1] = temp1[2];
	temp2[2] = temp1[1];
	temp2[3] = temp1[0];
	dw = *(DWORD*) &temp2;
}

void SwapEndian( WORD& w)
{
	BYTE	temp1[2],
			temp2[2];

	*(WORD*)&temp1 = w;
	temp2[0] = temp1[1];
	temp2[1] = temp1[0];
	w = *(WORD*) &temp2;
}

CCpkFile::CCpkFile()
{
}

CCpkFile::~CCpkFile()
{
}

void CCpkFile::OpenStreams(void)
{
	static const CCpkFile::SoundDescription	DefaultSndDesc = {0,1,8,0};		// 1 channel, 8 bits per sample, no compression
	static const CCpkFile::ChromaKey		DefaultChromakey = {0,0,0,0};	// No chromakeying.

	m_fileHdr.atomID = 'FILM';
	m_fileHdr.atomSize = 0;		// To be set later...
	m_fileHdr.version = '1.08';
	m_fileHdr.reserved = 0;

	m_frameDesc.atomID = 'FDSC';
	m_frameDesc.atomSize = sizeof(FrameDescription);
	m_frameDesc.cType = 'cvid';
	m_frameDesc.height = 0;					// To be set later...
	m_frameDesc.width = 0;					// To be set later...
	m_frameDesc.soundDesc = DefaultSndDesc;
	m_frameDesc.samplingRate = 0x56220000;	// Default:  22.050 kHz
	m_frameDesc.chromaKey = DefaultChromakey;

	m_sampleTable.atomID = 'STAB';
	m_sampleTable.atomSize = 0;		// To be set later...
	m_sampleTable.timeScale = 600;	// Default:  1/600th second.
	m_sampleTable.numSamples = 0;	// To be set later...

	m_sampleIndex = 0;
	m_sampleTime = 0;
	m_pSamples = NULL;
	m_remainingAudioTime = 0;
	m_remainingAudioBytes = 0;
}

void CCpkFile::CloseStreams(void)
{
	for ( DWORD i = 0 ; i < m_sampleTable.numSamples ; i++)
	{
		SwapEndian(m_pSamples[i].start);
		SwapEndian(m_pSamples[i].size);
		SwapEndian(m_pSamples[i].time);
		SwapEndian(m_pSamples[i].duration);
	}

	Seek( m_dwSampleTablePosition, begin);
	Write( m_pSamples, sizeof(FilmSample)*m_sampleTable.numSamples);
	free(m_pSamples);
	m_pSamples = NULL;
}

void CCpkFile::SetVideoInfo( DWORD numFrames, DWORD frameRate, DWORD frameScale, BITMAPINFO& bmi)
{
	m_frameDesc.height = bmi.bmiHeader.biHeight;
	m_frameDesc.width = bmi.bmiHeader.biWidth;

	m_frameRate = frameRate;
	m_frameScale = frameScale;
	m_sampleTable.numSamples += numFrames;
}

void CCpkFile::SetAudioInfo( DWORD numSamples, WAVEFORMATEX& wav)
{
	DWORD	nSamples,
			nSndByteSize = wav.nChannels * wav.wBitsPerSample/8 * wav.nSamplesPerSec / 4;

	if (numSamples)
	{
		nSamples = (numSamples + nSndByteSize-1) / nSndByteSize;
		m_sampleTable.numSamples += nSamples-1;
		m_frameDesc.soundDesc.reserved = 0x18;	// Saturn Cpk expects this?
		m_frameDesc.soundDesc.numChannels = (BYTE) wav.nChannels;
		m_frameDesc.soundDesc.numBitsSampled = (BYTE) wav.wBitsPerSample;
		m_frameDesc.soundDesc.compression = 0;	// No compression.
		m_frameDesc.samplingRate = (DWORD) wav.nSamplesPerSec << 16;
		m_remainingAudioBytes = numSamples;
	}
	else	// No audio stream.
	{
		m_frameDesc.soundDesc.reserved = 0x18;	// Saturn Cpk expects this?
		m_frameDesc.soundDesc.numChannels = 0;
		m_frameDesc.soundDesc.numBitsSampled = 0;
		m_frameDesc.soundDesc.compression = 0;	// No compression.
		m_frameDesc.samplingRate = 0;
	}
}
 
void CCpkFile::WriteHeader(void)
{
	BYTE			*	ptr;
						//reserved[16];
	FileHeader			fileHdr;
	FrameDescription	frameDesc;
	FilmSampleTable		sampleTable;

	m_sampleTable.atomSize = sizeof(FilmSampleTable) + m_sampleTable.numSamples*sizeof(FilmSample);
	m_fileHdr.atomSize = sizeof(FileHeader) + sizeof(FrameDescription)
		               + m_sampleTable.atomSize /*+ sizeof(reserved)*/;

	fileHdr = m_fileHdr;
	SwapEndian(fileHdr.atomID);
	SwapEndian(fileHdr.atomSize);
	SwapEndian(fileHdr.version);
	Write( &fileHdr, sizeof(fileHdr));

	frameDesc = m_frameDesc;
	SwapEndian(frameDesc.atomID);
	SwapEndian(frameDesc.atomSize);
	SwapEndian(frameDesc.cType);
	SwapEndian(frameDesc.height);
	SwapEndian(frameDesc.width);
	//SwapEndian(frameDesc.soundDesc);
	SwapEndian(frameDesc.samplingRate);
	//SwapEndian(frameDesc.chromaKey);
	Write( &frameDesc, sizeof(frameDesc));

	sampleTable = m_sampleTable;
	SwapEndian(sampleTable.atomID);
	SwapEndian(sampleTable.atomSize);
	SwapEndian(sampleTable.timeScale);
	SwapEndian(sampleTable.numSamples);
	Write( &sampleTable, sizeof(sampleTable));

	m_dwSampleTablePosition = GetPosition();

	m_pSamples = (FilmSample*) malloc( sizeof(FilmSample)*(m_sampleTable.numSamples+1));
	ASSERT(m_pSamples);
	memset( m_pSamples, 0, sizeof(FilmSample)*m_sampleTable.numSamples);
	Write( m_pSamples, sizeof(FilmSample)*m_sampleTable.numSamples);

	// First audio sample, if there is an audio stream.
	if ( m_frameDesc.soundDesc.numChannels)
	{
		m_pSamples[0].size =
			(((m_frameDesc.soundDesc.numChannels*m_frameDesc.soundDesc.numBitsSampled/8)
			* (m_frameDesc.samplingRate/2)) >> 16) & ~3;
		m_pSamples[0].duration = 1;
		m_pSamples[0].time = 0xffffffff;
		m_sampleIndex = 1;
		m_remainingAudioTime = m_sampleTable.timeScale/2;
		m_remainingAudioBytes -= m_pSamples[0].size;

		ptr = (BYTE*) malloc(m_pSamples[0].size);
		ASSERT(ptr);
		memset( ptr, 0, m_pSamples[0].size);
		Write( ptr, m_pSamples[0].size);
		free(ptr);
	}

//	memset( reserved, 0, sizeof(reserved));
//	Write( reserved, sizeof(reserved));
};


void CCpkFile::WriteVideoStream( char* pBuf, DWORD cBytes)
{
	char	*	pDst,
			*	ptr;
	WORD		i,
				j,
				numTags;
	DWORD		hdrSize,
				hdrPadding,
				tag,
				tagSize,
				tagPadding,
			*	pTag,
				subtag,
				subtagSize,
				subtagSizePadded;

	pDst = (char*) malloc(cBytes*2);
	ASSERT(pDst);
	ptr = pDst;

	hdrSize = *(DWORD*)pBuf;
	numTags = *((WORD*)pBuf+4);
	SwapEndian(hdrSize);
	SwapEndian(numTags);
	memcpy( ptr, pBuf, sizeof(DWORD)+3*sizeof(WORD));
	ptr += sizeof(DWORD) + 3*sizeof(WORD);
	pBuf += sizeof(DWORD) + 3*sizeof(WORD);
	memset( ptr, 0, sizeof(WORD));
	ptr += sizeof(WORD);
	
	for ( i = 0, hdrPadding = 0 ; i < numTags ; i++)
	{
		pTag = (DWORD*)ptr;
		tag = *(DWORD*)pBuf;
		SwapEndian(tag);
		tagSize = tag & 0x00ffffff;
		memcpy( ptr, pBuf, 3*sizeof(DWORD));
		ptr += 3*sizeof(DWORD);
		pBuf += 3*sizeof(DWORD);

		for ( j = 0, tagPadding = 0 ; j < 3 ; j++)
		{
			subtag = *(DWORD*)pBuf;
			pBuf += sizeof(DWORD);
			SwapEndian(subtag);
			subtagSize = subtag & 0x00ffffff;
			subtagSizePadded = (subtagSize + 3) & ~3;
			subtag = (subtag & 0xff000000) | subtagSizePadded;
			SwapEndian(subtag);
			*(DWORD*)ptr = subtag;
			ptr += sizeof(DWORD);
			memcpy( ptr, pBuf, subtagSize-sizeof(DWORD));
			ptr += subtagSize - sizeof(DWORD);
			pBuf += subtagSize - sizeof(DWORD);
			memset( ptr, 0, subtagSizePadded-subtagSize);
			ptr += subtagSizePadded - subtagSize;

			tagPadding += subtagSizePadded - subtagSize;
		}

		tagSize += tagPadding;
		tag = (tag & 0xff000000) | tagSize;
		SwapEndian(tag);
		*pTag = tag;

		hdrPadding += tagPadding;
	}

	cBytes += sizeof(WORD) + hdrPadding;
	hdrSize = cBytes - 8;
	SwapEndian(hdrSize);
	SwapEndian(numTags);
	*(DWORD*)pDst = hdrSize;

	Write( pDst, cBytes);
	free(pDst);

	if ( m_sampleIndex > 0)
		m_pSamples[m_sampleIndex].start =
			m_pSamples[m_sampleIndex-1].start + m_pSamples[m_sampleIndex-1].size;
	else
		m_pSamples[0].start = 0;
	m_pSamples[m_sampleIndex].size = cBytes;
	m_pSamples[m_sampleIndex].time = m_sampleTime;
	m_pSamples[m_sampleIndex].duration = (m_sampleTable.timeScale*m_frameScale + m_frameRate/2)/m_frameRate;
	m_sampleTime += m_pSamples[m_sampleIndex].duration;
	m_sampleIndex++;

	// Interleave empty audio frames, if there is an audio stream.
	if ( m_frameDesc.soundDesc.numChannels && m_remainingAudioBytes > 0)
	{
		m_remainingAudioTime -= m_pSamples[m_sampleIndex-1].duration;
		if ( m_remainingAudioTime <= m_sampleTable.timeScale*3/8)
		{
			m_pSamples[m_sampleIndex].start =
				m_pSamples[m_sampleIndex-1].start + m_pSamples[m_sampleIndex-1].size;
			m_pSamples[m_sampleIndex].size =
				(((m_frameDesc.soundDesc.numChannels*m_frameDesc.soundDesc.numBitsSampled/8)
				* (m_frameDesc.samplingRate/4)) >> 16) & ~3;
			m_pSamples[m_sampleIndex].time = 0xffffffff;
			m_pSamples[m_sampleIndex].duration = 1;

			if ( m_remainingAudioBytes < m_pSamples[m_sampleIndex].size)
			{
				//m_pSamples[m_sampleIndex].size = (m_remainingAudioBytes + 3) & ~3;
				m_remainingAudioBytes = 0;
			}
			else
				m_remainingAudioBytes -= m_pSamples[m_sampleIndex].size;

			pDst = (char*) malloc(m_pSamples[m_sampleIndex].size);
			ASSERT(pDst);
			memset( pDst, 0, m_pSamples[m_sampleIndex].size);
			Write( pDst, m_pSamples[m_sampleIndex].size);
			free(pDst);

			m_remainingAudioTime += m_sampleTable.timeScale/4;
			m_sampleIndex++;
		}
	}
}

void CCpkFile::WriteAudioStream( char* pBuf, DWORD cBytes)
{
	char	*	ptr;
	DWORD		i,
				temp = 0,
				dwDataPosition = m_dwSampleTablePosition
				               + m_sampleTable.numSamples*sizeof(FilmSample)
							   /*+ 16*/;
	
	if ( m_frameDesc.soundDesc.numChannels == 0)
		return;

	for ( i = 0, ptr = pBuf ; i < cBytes ; i++, ptr++)
		*ptr -= (char)0x80;

	for ( i = 0 ; cBytes > 0 ; i++)
	{
		if ( m_pSamples[i].time == 0xffffffff)
		{
			Seek( dwDataPosition, begin);
			if ( m_pSamples[i].size > cBytes)
			{
				Write( pBuf, cBytes);
				ptr = (char*) malloc( m_pSamples[i].size-cBytes);
				memset( ptr, 0, m_pSamples[i].size-cBytes);
				Write( ptr, m_pSamples[i].size-cBytes);
				free(ptr);
				cBytes = 0;
			}
			else
			{
				Write( pBuf, m_pSamples[i].size);
				pBuf += m_pSamples[i].size;
				cBytes -= m_pSamples[i].size;
			}
		}
		dwDataPosition += m_pSamples[i].size;
	}
}
